import math
import numpy as np
import heapq
import geopy.distance
import tripy # Needs to be rewrited for the triangulation function

from .const import *
from .msg import *
from .common import *
from .vector import *
from .relation import *

def ptXY2LatLonMercator(
    ptXY:       "Point in (x, y) coordinates"
    ) -> "Given a point in (x, y), try to map it with a (lat, lon) coordinate with necessary inputs":
    # Ref: https://wiki.openstreetmap.org/wiki/Mercator#Python
    (x, y) = ptXY
    lon = math.degrees(y / CONST_EARTH_RADIUS_METERS)
    lat = math.degrees(2 * math.atan(math.exp(x / CONST_EARTH_RADIUS_METERS)) - math.pi / 2.0)
    ptLatLon = (lat, lon)
    return ptLatLon

def ptLatLon2XYMercator(
    ptLatLon:   "Point in (lat, lon) coordinates" = None
    ) -> "Given a point int (lat, lon), try to map it with a (x, y) coordinates with necessary inputs":
    # Ref: https://wiki.openstreetmap.org/wiki/Mercator#Python
    (lat, lon) = ptLatLon
    y = math.radians(lon) * CONST_EARTH_RADIUS_METERS
    x = math.log(math.tan(math.pi / 4 + math.radians(lat) / 2)) * CONST_EARTH_RADIUS_METERS
    ptXY = (x, y)
    return ptXY

def twMovingPtInsidePolyXY(
    ptXY:       "Point that are moving" = None,
    polyXY:     "Polygon that are moving" = None,
    vecXYPt:    "Moving speed vector of the point in XY" = None,
    vecXYPoly:  "Moving speed vector of the polygon in XY" = None,
    ) -> "Given a moving point in (x, y), a moving polygon in Euclidean2D, returns \
        1) the time window of when the point is inside the polygon, or \
        2) empty list if the point is not going to be inside the polygon":

    # Initialize ==============================================================
    tw = []
    # Convert the speed of both objects into the speed of pt
    vecXY = calXYVecSubtract(vecXYPt, vecXYPoly)

    # Move axis, so that ptXY = (0, 0) ========================================
    polyOffSet = []
    for p in polyXY:
        polyOffSet.append((p[0] - ptXY[0], p[1] - ptXY[1]))
    rayOffset = [(0, 0), vecXY]

    # Check if the point is going to go into the polygon ======================
    intersectFlag = isRayCrossPoly(rayOffset, polyOffSet)
    if (not intersectFlag):
        return []

    # Find the intersecting pts of the ray and all edges ======================
    # Polygon might not be convex, there could be multiple time windows
    intPts = []
    for i in range(-1, len(polyOffSet) - 1):
        ptInt = intSeg2Ray([polyOffSet[i], polyOffSet[i + 1]], rayOffset)
        if (ptInt != None):
            intPts.append(ptInt)

    # Sort intersection points by dist ========================================
    # Convert intPts to nodes
    nodes = {}
    nodes[0] = {'loc': (0, 0), 'color': 'blue'}
    for i in range(len(intPts)):
        nodes[i + 1] = {'loc': (intPts[i][0], intPts[i][1]), 'color': 'red'}
    sortedSeq = getSortedNodesByDist(
        nodes = nodes,
        edges = 'Euclidean',
        refNodeID = 0)

    # Calculate relative distances to ptXY ====================================
    dist = []
    for i in range(len(sortedSeq)):
        dist.append(distEuclidean2D((0, 0), nodes[sortedSeq[i]]['loc']))
    timeStamp = []
    absVXY = math.sqrt(vecXY[0]**2 + vecXY[1]**2)
    for i in dist:
        timeStamp.append(i / absVXY)

    # Confirm time windows ====================================================
    for i in range(len(dist) - 1):
        mid = ((nodes[sortedSeq[i]]['loc'][0] / 2 + nodes[sortedSeq[i + 1]]['loc'][0] / 2), 
            (nodes[sortedSeq[i]]['loc'][1] / 2 + nodes[sortedSeq[i + 1]]['loc'][1] / 2))
        if (isPtInsidePoly(mid, polyOffSet)):
            tw.append([timeStamp[i], timeStamp[i + 1]])

    return tw

def twMovingPtInsidePolyLatLon(
    ptLatLon:      "Point that are moving" = None,
    polyLatLon:    "Polygon that are moving" = None,
    vecPolarPt:    "Moving speed vector of the point in XY" = None,
    vecPolarPoly:  "Moving speed vector of the polygon in XY" = None,
    ) -> "Given a moving point in (lat, lon), a moving polygon, returns \
        1) the time window of when the point will be inside the polygon, or \
        2) None if not going to be inside the polygon":

    # Steps ===================================================================
    # 1. Project the Lat/Lon into XY space, using Mercator Projection
    # 2. Calculate the `twMovingPtInsidePolyXY()`, find the intersections
    # 3. Convert the intersections back to (lat, lon)
    # 4. Calculated the time windows based on speed vector in LatLon

    # Initialize ==============================================================
    tw = []
    ptXY = ptLatLon2XYMercator(ptLatLon)
    polyXYOffset = []
    for p in polyLatLon:
        pXY = ptLatLon2XYMercator(p)
        polyXYOffset.append((pXY[0] - ptXY[0], pXY[1] - ptXY[1]))
    vecXY = calXYVecSubtract(
        vecPolar2XY(vecPolarPt), 
        vecPolar2XY(vecPolarPoly))
    rayXYOffset = [(0, 0), vecXY]

    # Find intersection in XY space ===========================================
    intersectFlag = isRayCrossPoly(rayXYOffset, polyXYOffset)
    # print(intersectFlag)
    if (not intersectFlag):
        return []
    intPtsXY = []
    for i in range(-1, len(polyXYOffset) - 1):
        ptIntXY = intSeg2Ray([polyXYOffset[i], polyXYOffset[i + 1]], rayXYOffset)
        if (ptIntXY != None):
            intPtsXY.append(ptIntXY)

    # Convert those points back to (lat, lon) =================================
    intPtsLatLon = []
    for p in intPtsXY:
        pXY = (p[0] + ptXY[0], p[1] + ptXY[1])
        intPtsLatLon.append(ptXY2LatLonMercator(pXY))
    # print(intPtsLatLon)

    # Sort intersection points by dist ========================================
    # Convert intPts to nodes
    intPtsLatLonDict = {}
    intPtsLatLonDict[0] = {'loc': (ptLatLon[0], ptLatLon[1])}
    for i in range(len(intPtsLatLon)):
        intPtsLatLonDict[i + 1] = {'loc': (intPtsLatLon[i][0], intPtsLatLon[i][1])}
    sortedSeq = getSortedNodesByDist(
        nodes = intPtsLatLonDict,
        edges = 'LatLon',
        refNodeID = 0)

    # Calculate relative distances to ptLatLon ======================================
    dist = []
    for i in range(len(sortedSeq)):
        dist.append(distLatLon(ptLatLon, intPtsLatLonDict[sortedSeq[i]]['loc'], distUnit='meter'))
    timeStamp = []
    vecPolar = vecXY2Polar(vecXY)
    absVXY = vecPolar[0]
    for i in dist:
        timeStamp.append(i / absVXY)

    # Confirm time windows ====================================================
    for i in range(len(dist) - 1):
        mid = ((intPtsLatLonDict[sortedSeq[i]]['loc'][0] / 2 + intPtsLatLonDict[sortedSeq[i + 1]]['loc'][0] / 2), 
            (intPtsLatLonDict[sortedSeq[i]]['loc'][1] / 2 + intPtsLatLonDict[sortedSeq[i + 1]]['loc'][1] / 2))
        if (isPtInsidePoly(mid, polyLatLon)):
            tw.append([timeStamp[i], timeStamp[i + 1]])

    return tw

def distEuclidean2D(
    pt1:        "First coordinate, in (x, y)", 
    pt2:        "Second coordinate, in (x, y)"
    ) -> "Gives a Euclidean distance based on two coords, if two coordinates are the same, return a small number":

    # 2-Norm ==================================================================
    if (pt1 != None and pt2 != None):
        return math.sqrt((pt1[0] - pt2[0]) ** 2 + (pt1[1] - pt2[1]) ** 2)
    else:
        return 0

def distLatLon(
    pt1:        "First coordinate, in (lat, lon)", 
    pt2:        "Second coordinate, in (lat, lon)",
    distUnit:   "Unit of distance\
                 1) String (default) 'mile'\
                 2) String 'meter'\
                 3) String 'kilometer'" = 'mile'
    ) -> "Gives a Euclidean distance based on two lat/lon coords, if two coordinates are the same, return a small number":
    
    # Get radius as in distUnit ===============================================
    R = None
    if (distUnit == 'mile'):
        R = CONST_EARTH_RADIUS_MILES
    elif (distUnit == 'meter'):
        R = CONST_EARTH_RADIUS_METERS
    elif (distUnit == 'kilometer'):
        R = CONST_EARTH_RADIUS_METERS / 1000
    else:
        print("ERROR: Unrecognized distance unit, options are 'mile', 'meter', 'kilometer'")
        return

    # Calculate distance ======================================================
    if (pt1 != None and pt2 != None):
        (lat1, lon1) = pt1
        (lat2, lon2) = pt2
        phi1, phi2 = math.radians(lat1), math.radians(lat2) 
        dphi = math.radians(lat2 - lat1)
        dlambda = math.radians(lon2 - lon1)
        a = math.sin(dphi / 2) ** 2 + math.cos(phi1) * math.cos(phi2) * math.sin(dlambda / 2) ** 2
        return 2 * R * math.atan2(math.sqrt(a), math.sqrt(1 - a))
    else:
        return CONST_EPSILON

def distPt2Line(
    pt:         "2-tuple of coordinate (x, y)",
    line:       "List of 2 pts which defines a line"
    ) -> "Distance from a given point to a given line":

    # Validation ==============================================================
    if (is2PtsSame(line[0], line[1])):
        print(ERROR_ZERO_VECTOR)
        return None
    
    # Initialize ==============================================================
    ptA = line[0]
    ptB = line[1]
    ptS = pt

    # Check if the pt is on line, if so the distance is 0 =====================
    if (isPtOnLine(pt, line)):
        return 0.0

    # Calculate dist to line ==================================================
    areaSAB = calTriangleAreaByCoords(ptS, ptA, ptB)
    bottom = distEuclidean2D(ptA, ptB)
    height = 2 * areaSAB / bottom
    dist = height

    return dist

def distPt2Seg(
    pt:         "2-tuple of coordinate (x, y)",
    seg:        "List of 2 pts which defines a line segment"
    ) -> "Distance from a given point to a given line segment":
    
    # Validation ==============================================================
    if (is2PtsSame(line[0], line[1])):
        print(ERROR_ZERO_VECTOR)
        return None

    # Initialize ==============================================================
    ptA = line[0]
    ptB = line[1]
    ptS = pt

    # Check if the pt is on line, if so the distance is 0 =====================
    if (isPtOnLine(pt, line)):
        return 0.0

    # Rays ====================================================================
    rayAS = [ptS[0] - ptA[0], ptS[1] - ptA[1]]
    rayAB = [ptB[0] - ptA[0], ptB[1] - ptA[1]]
    rayBS = [ptS[0] - ptB[0], ptS[1] - ptB[1]]
    rayBA = [ptA[0] - ptB[0], ptA[1] - ptB[1]]

    # cos value for A angle and B angle =======================================
    cosSAB = cosRay2Ray(rayAS, rayAB)
    cosSBA = cosRay2Ray(rayBS, rayBA)

    # Calculate dist to line ==================================================
    # if both angles are sharp, the closest point will be in the line, otherwise the closest point is at the edge
    if (cosSAB >= 0 and cosSBA >= 0):
        areaSAB = calTriangleAreaByCoords(ptS, ptA, ptB)
        bottom = distEuclidean2D(ptA, ptB)
        height = 2 * areaSAB / bottom
        dist = height
    else:
        distAS = distEuclidean2D(ptS, ptA)
        distBS = distEuclidean2D(ptS, ptB)
        dist = min(distAS, distBS)

    return dist

def getTauEuclidean(
    nodes:      "Dictionary, returns the coordinate of given nodeID, \
                    {\
                        nodeID1: {'loc': (x, y)}, \
                        nodeID2: {'loc': (x, y)}, \
                        ... \
                    }" = None,
    nodeIDs:    "1) String (default) 'All', or \
                 2) A list of node IDs" = 'All',
    speed:      "Ratio: Dist / Time" = 1
    ) -> "Dictionary, {(nodeID1, nodeID2): dist, ...}":

    # Define nodeIDs ==========================================================
    if (type(nodeIDs) is not list):
        if (nodeIDs == 'All'):
            nodeIDs = []
            for i in nodes:
                nodeIDs.append(i)

    # Get tau =================================================================
    tau = {}
    for i in nodeIDs:
        for j in nodeIDs:
            if (i != j):
                t = distEuclidean2D(nodes[i]['loc'], nodes[j]['loc']) / speed
                tau[i, j] = t
                tau[j, i] = t
            else:
                tau[i, j] = CONST_EPSILON
                tau[j, i] = CONST_EPSILON
    return tau

def getTauLatLon(
    nodes:      "Dictionary, returns the coordinate of given nodeIDs, \
                    {\
                        nodeIDs1: {'loc': (x, y)}, \
                        nodeIDs2: {'loc': (x, y)}, \
                        ... \
                    }" = None,
    nodeIDs:    "1) String (default) 'All', or \
                 2) A list of node IDs" = 'All',                    
    distUnit:   "Unit of distance\
                 1) String (default) 'mile'\
                 2) String 'meter'\
                 3) String 'kilometer'" = 'mile',
    speed:      "Ratio: Dist / Time" = 1
    ) -> "Dictionary, {(nodeID1, nodeID2): dist, ...}":

    # Define nodeIDs ==========================================================
    if (type(nodeIDs) is not list):
        if (nodeIDs == 'All'):
            nodeIDs = []
            for i in nodes:
                nodeIDs.append(i)

    # Get tau =================================================================
    tau = {}
    for i in nodeIDs:
        for j in nodeIDs:
            if (i != j):
                t = distLatLon(nodes[i]['loc'], nodes[j]['loc'], distUnit) / speed
                tau[i, j] = t
                tau[j, i] = t
            else:
                tau[i, j] = CONST_EPSILON
                tau[j, i] = CONST_EPSILON
    return tau

def getPerpendicularLine(
    pt:         "Point which the line will go through",
    vec:        "The vector that perpendicular to the line"
    ) -> "Given a point, a vector, returns a line that is going through this point and perpendicular to the vector":

    # Get the direction =======================================================
    heading = getHeadingXY(pt, (pt[0] + vec[0], pt[1] + vec[1]))
    newHeading1 = heading + 90
    newHeading2 = heading - 90
    while (newHeading1 > 360):
        newHeading1 -= 360
    while (newHeading2 > 360):
        newHeading2 -= 360
    while (newHeading1 < 0):
        newHeading1 += 360
    while (newHeading2 < 0):
        newHeading2 += 360

    # Get line ================================================================
    line = [pointInDistXY(pt, newHeading1, 10), pointInDistXY(pt, newHeading2, 10)]

    return line

def getHeadingXY(
    pt1:        "Current location", 
    pt2:        "Targeted location"
    ) -> "Given current location and a goal location, calculate the heading. Up is 0-degrees, clock-wise":
    
    vec = (pt2[0] - pt1[0], pt2[1] - pt1[1])
    (_, vDeg) = vecXY2Polar(vec)

    return vDeg

def getHeadingLatLon(
    pt1:        "Current location as in [lat, lon]", 
    pt2:        "Target location as in [lat, lon]"
    ) -> "Given current location and a goal location, calculate the heading. Up is 0-degrees, clock-wise":
    # Ref: https://github.com/manuelbieh/Geolib/issues/28
    (lat1, lon1) = pt1
    (lat2, lon2) = pt2
    lat1 = math.radians(lat1)
    lon1 = math.radians(lon1)
    lat2 = math.radians(lat2)
    lon2 = math.radians(lon2)
    
    dLon = lon2 - lon1
    if abs(dLon) > math.pi:
        if dLon > 0.0:
            dLon = -(2.0 * math.pi - dLon)
        else:
            dLon = (2.0 * math.pi + dLon)

    tan1 = math.tan(lat1 / 2.0 + math.pi / 4.0)
    tan2 = math.tan(lat2 / 2.0 + math.pi / 4.0)
    phi = math.log(tan2 / tan1)
    deg = (math.degrees(math.atan2(dLon, phi)) + 360.0) % 360.0
    
    return deg

def getRndPtOnNetworkLatLon(
    network:    "List of dictionary of networks in the format of \
                {\
                    'start': (lat, lon),\
                    'end': (lat, lon),\
                }" = None,
    ):

    # Calculate the length of each edge =======================================
    lengths = []
    for edge in network:
        lengths.append(distLatLon(edge['start'], edge['end']))
    idx = rndPick(lengths)

    # Randomly generate a point within the egde selected ======================
    edgeLength = lengths[idx]
    edgeDist = random.uniform(0, 1) * edgeLength

    # Get location ============================================================
    getHeadingLatLon(network[idx][0], network[idx][1])
    (lat, lon) = pointInDistLatLon(network[idx][0], edgeDist)

    return (lat, lon)

def getRndPtUniformSquare(
    xRange:    "The range of x coordinates" = (0, 100),
    yRange:    "The range of y coordinates" = (0, 100)
    ) -> "Given the range of x, y, returns a random point in the square defined by the ranges":
    x = random.randrange(xRange[0], xRange[1])
    y = random.randrange(yRange[0], yRange[1])
    return (x, y)

def getRndPtUniformTriangle(
    triangle:   "The triangle for generating random points" = None
    ) -> "Given a triangle, generate a random point in the triangle uniformly":
    
    # Get three extreme points ================================================
    [x1, y1] = triangle[0]
    [x2, y2] = triangle[1]
    [x3, y3] = triangle[2]

    # Generate random points ==================================================
    rndR1 = random.uniform(0, 1)
    rndR2 = random.uniform(0, 1)
    x = (1 - math.sqrt(rndR1)) * x1 + math.sqrt(rndR1) * (1 - rndR2) * x2 + math.sqrt(rndR1) * rndR2 * x3
    y = (1 - math.sqrt(rndR1)) * y1 + math.sqrt(rndR1) * (1 - rndR2) * y2 + math.sqrt(rndR1) * rndR2 * y3

    return (x, y)

def getRndPtUniformPoly(
    poly:       "The polygon for generating random points" = None
    ) -> "Given a polygon, generate a random point in the polygons uniformly":

    # Get list of triangles ===================================================
    lstTriangle.extend(tripy.earclip(poly))

    # Weight them and make draws ==============================================
    lstWeight = []
    for i in range(len(lstTriangle)):
        lstWeight.append(calTriangleAreaByCoords(lstTriangle[i][0], lstTriangle[i][1], lstTriangle[i][2]))

    # Select a triangle and randomize a point in the triangle =================
    idx = rndPick(lstWeight)
    (x, y) = getRndPtUniformTriangle(lstTriangle[idx])

    return (x, y)

def getRndPtUniformPolys(
    polys:       "A list of polygons for generating random points" = None
    ) -> "Given a list of polygons, generate a random point in the polygons uniformly":

    # Get all triangulated triangles ==========================================
    lstTriangle = []
    for p in polys:
        lstTriangle.extend(tripy.earclip(p))

    # Weight them and make draws ==============================================
    lstWeight = []
    for i in range(len(lstTriangle)):
        lstWeight.append(calTriangleAreaByCoords(lstTriangle[i][0], lstTriangle[i][1], lstTriangle[i][2]))

    # Select a triangle and randomize a point in the triangle =================
    idx = rndPick(lstWeight)
    (x, y) = getRndPtUniformTriangle(lstTriangle[idx])

    return (x, y)

def getRndPtCluster():

    return

def getRndPtUniformCircleXY():

    return

def getRndPtUniformCircleLatLon():

    return

def getRndPtNormalCircleXY():

    return

def getRndPtNormalCircleLatLon():

    return

# [Temp]
def tempAndPleaseDoNotCallMe(
    distr:      "Spatial distribution of nodes, the options are \
                 1) String, (default) 'uniformSquare', or \
                 2) String, 'uniformCircle', or \
                 3) String, 'uniformPoly', or\
                 4) String, (not available) 'uniformOnNetwork', or\
                 5) String, 'clustered, or \
                 6) String, (not available) 'normal2D'" = 'uniformSquare',
    distrArgs:  "Dictionary that describes the distribution of the nodes\
                 1) for 'uniformSquare'\
                    {\
                        'xRange': A 2-tuple with minimum/maximum range of x, default as (0, 100), \
                        'yRange': A 2-tuple with minimum/maximum range of y, default as (0, 100), \
                    }\
                 2) for 'uniformCircle'\
                    {\
                        'centerLoc': centering location \
                        'radius': radius of the circle \
                    }\
                 3) for 'uniformPoly'\
                    {\
                        'poly': polygon of the area, (no holes)\
                        'polys': list of polygons \
                    }\
                 4) for 'uniformOnNetwork'\
                    {\
                        'network': list of arcs that can be sampled \
                    }\
                 5) for 'clustered\
                    {\
                        'numCluster': number of cluster centers\
                        'xRange': xRange of cluster centroid\
                        'yRange': yRange of cluster centroid\
                        'poly': polygon for customers\
                        'centroidLocs': list of cluster center locations\
                        'clusterDiameter': the spread of nodes, in diameter\
                    }\
                 6) for 'normal2D'\
                " = {
                    'xRange': (0, 100),
                    'yRange': (0, 100)
                 }
    ) -> "Given a distribution and the arguments, return a random point in (x, y)":
    # Generate instance =======================================================
    if (distr == "uniformSquare"):
        # Sanity check --------------------------------------------------------
        if (distrArgs == None):
            distrArgs = {
                'xRange': (0, 100),
                'yRange': (0, 100)
            }
        if ('xRange' not in distrArgs):
            distrArgs['xRange'] = (0, 100)
        if ('yRange' not in distrArgs):
            distrArgs['yRange'] = (0, 100)
        # Create nodes --------------------------------------------------------
        for n in nodeIDs: 
            x = random.randrange(distrArgs['xRange'][0], distrArgs['xRange'][1])
            y = random.randrange(distrArgs['yRange'][0], distrArgs['yRange'][1])
            nodes[n] = {'loc': (x, y)}
    elif (distr == "uniformCircle"):
        # Sanity check --------------------------------------------------------
        if (distrArgs == None):
            distrArgs = {
                'centerLoc': [0, 0],
                'radius': 100
            }
        if ('centerLoc' not in distrArgs or 'radius' not in distrArgs):
            print(ERROR_MISSING_DISTRARGS_UNICC)
            return
        # Create nodes --------------------------------------------------------
        for n in nodeIDs:
            theta = random.uniform(0, 2 * math.pi)
            r = math.sqrt(random.uniform(0, distrArgs['radius'] ** 2))
            x = distrArgs['centerLoc'][0] + r * math.cos(theta)
            y = distrArgs['centerLoc'][1] + r * math.sin(theta)
            nodes[n] = {'loc': (x, y)}
    elif (distr == "uniformPoly"):
        # Sanity check --------------------------------------------------------
        if (distrArgs == None):
            print(ERROR_MISSING_DISTRARGS)
            return
        if ('poly' not in distrArgs and 'polys' not in distrArgs):
            print(ERROR_MISSING_DISTRARGS_UNIPOLY)
            return
        # Create nodes --------------------------------------------------------
        polys = []
        if ('poly' in distrArgs and 'polys' not in distrArgs):
            polys = [distrArgs['poly']]
        elif ('poly' in distrArgs and 'polys' in distrArgs):
            polys = [distrArgs['poly']]
            polys.extend(distrArgs['polys'])
        elif ('polys' in distrArgs):
            polys = distrArgs['polys']
        # Get all triangulated triangles
        lstTriangle = []
        for p in polys:
            lstTriangle.extend(tripy.earclip(p))
        # Weight them and make draws
        lstWeight = []
        for i in range(len(lstTriangle)):
            lstWeight.append(calTriangleAreaByCoords(lstTriangle[i][0], lstTriangle[i][1], lstTriangle[i][2]))
        for n in nodeIDs:
            idx = rndPick(lstWeight)
            [x1, y1] = lstTriangle[idx][0]
            [x2, y2] = lstTriangle[idx][1]
            [x3, y3] = lstTriangle[idx][2]
            rndR1 = np.random.uniform(0, 1)
            rndR2 = np.random.uniform(0, 1)
            rndX = (1 - math.sqrt(rndR1)) * x1 + math.sqrt(rndR1) * (1 - rndR2) * x2 + math.sqrt(rndR1) * rndR2 * x3
            rndY = (1 - math.sqrt(rndR1)) * y1 + math.sqrt(rndR1) * (1 - rndR2) * y2 + math.sqrt(rndR1) * rndR2 * y3
            nodes[n] = {'loc': (rndX, rndY)}
    elif (distr == "uniformOnNetwork"):
        print("Stay tune")
        return
    elif (distr == "clustered"):
        # Sanity check --------------------------------------------------------
        if (distrArgs == None):
            print(ERROR_MISSING_DISTRARGS)
            return
        if (('numCluster' not in distrArgs or 'xRange' not in distrArgs or 'yRange' not in distrArgs) and 'centroidLocs' not in distrArgs):
            print(ERROR_MISSING_DISTRARGS_CLUSTER)
            return
        if ('clusterDiameter' not in distrArgs):
            distrArgs['clusterDiameter'] = 20
        # Create nodes --------------------------------------------------------
        centroidLocs = []
        if ('centroidLocs' in distrArgs):
            centroidLocs = distrArgs['centroidLocs']
        else:
            for cl in range(distrArgs['numCluster']):
                x = random.randrange(distrArgs['xRange'][0], distrArgs['xRange'][1])
                y = random.randrange(distrArgs['yRange'][0], distrArgs['yRange'][1])
                centroidLocs.append([x, y])
        for n in nodeIDs:
            # First pick a centroidLoc
            idx = random.randint(0, len(centroidLocs) - 1)
            ctrLoc = centroidLocs[idx]
            theta = random.uniform(0, 2 * math.pi)
            r = math.sqrt(random.uniform(0, distrArgs['clusterDiameter']))
            x = ctrLoc[0] + r * math.cos(theta)
            y = ctrLoc[1] + r * math.sin(theta)
            nodes[n] = {'loc': (x, y)}
    elif (distr == "normal2D"):
        print("Stay tune")
        return
    else:
        print(ERROR_INCOR_DISTARG)
        return

    return rndPtXY

def rectInWidthLengthOrientationXY(
    centroidXY: "Centroid of rectangular in (x, y)" = None,
    width:      "Width of the rectangular" = None,
    length:     "Length of the rectangular" = None,
    oriDeg:     "Orientation of the rectangular" = None
    ) -> "Given args for the width, length, and orientation of rectangular, returns the coordinates in (x, y)":

    # Create four corner points ===============================================
    ptTemp1 = pointInDistXY(centroidXY, oriDeg, length / 2)
    pt1 = pointInDistXY(ptTemp1, oriDeg + 90, width / 2)
    pt2 = pointInDistXY(ptTemp1, oriDeg - 90, width / 2)
    ptTemp2 = pointInDistXY(centroidXY, oriDeg + 180, length / 2)
    pt3 = pointInDistXY(ptTemp2, oriDeg + 90, width / 2)
    pt4 = pointInDistXY(ptTemp2, oriDeg - 90, width / 2)

    # Get the rectangular =====================================================
    rect = [pt1, pt2, pt4, pt3]

    return rect

def rectInWidthLengthOrientationLatLon(
    centroidLatLon: "Centroid of rectangular in (lat, lon)" = None,
    widthInMeter: "Width of the rectangular in meters" = None,
    lengthInMeter: "Length of the rectangular in meters" = None,
    oriDeg:     "Orientation of the rectangular" = None
    ) -> "Given args for the width, length, and orientation of rectangular, returns the coordinates in (lat, lon)":

    # Create four corner points ===============================================
    ptTemp1 = pointInDistLatLon(centroidLatLon, oriDeg, lengthInMeter / 2)
    pt1 = pointInDistLatLon(ptTemp1, oriDeg + 90, widthInMeter / 2)
    pt2 = pointInDistLatLon(ptTemp1, oriDeg - 90, widthInMeter / 2)
    ptTemp2 = pointInDistLatLon(centroidLatLon, oriDeg + 180, lengthInMeter / 2)
    pt3 = pointInDistLatLon(ptTemp2, oriDeg + 90, widthInMeter / 2)
    pt4 = pointInDistLatLon(ptTemp2, oriDeg - 90, widthInMeter / 2)

    # Get the rectangular =====================================================
    rect = [pt1, pt2, pt4, pt3]

    return rect

def getRndPolyXY(

    ) -> "Give a centroid of poly and some arguments, return a randomized polygon in (x, y)":

    return rndPolyXY

def getRndPolyLatLon(

    ) -> "Give a centroid of poly and some arguments, return a randomized polygon in (lat, lon)":

    return rndPolyLatLon

def pointInDistXY(
    pt:         "Starting location" = None, 
    direction:  "Direction towards the destination, Up is 0-degrees, clock-wise" = None, 
    dist:       "Distance from origin location to destination location" = None
    ) -> "A location in distance with given direction, in [lat, lon] form.":
    x = pt[0] + dist * math.sin(math.radians(direction))
    y = pt[1] + dist * math.cos(math.radians(direction))
    return (x, y)

def pointInDistLatLon(
    pt:         "Starting location" = None, 
    direction:  "Direction towards the destination, North is 0-degree, East is 90-degrees" = None, 
    distMeters: "Distance from origin location to destination location" = None
    ) -> "A location in distance with given direction, in [lat, lon] form.":
    newLoc = list(geopy.distance.distance(meters=distMeters).destination(point=pt, bearing=direction))[:2]
    return newLoc

def getNeighborCluster(
    nodes:      "Dictionary, returns the coordinate of given nodeID, \
                    {\
                        nodeID1: {'loc': (x, y)}, \
                        nodeID2: {'loc': (x, y)}, \
                        ... \
                    }" = None, 
    edges:      "1) String (default) 'Euclidean' or \
                 2) String 'LatLon'" = "Euclidean",
    nodeIDs:    "1) String (default) 'All', or \
                 2) A list of node IDs" = 'All',
    diameter:   "The diameter of the neighborhood" = None
    ) -> "Given a dictionary of locations, an a radius, return the sets that any two locations are within the distance":

    # Define nodeIDs ==========================================================
    if (type(nodeIDs) is not list):
        if (nodeIDs == 'All'):
            nodeIDs = []
            for i in nodes:
                nodeIDs.append(i)

    # Define edges ============================================================
    if (edges == 'Euclidean'):
        edges = getTauEuclidean(nodes)
    elif (edges == 'LatLon'):
        edges = getTauLatLon(nodes)
    else:
        print("Error: Incorrect type `edges`")
        return None

    # Initialize ==============================================================
    seedClique = []
    # For each node get the neighbors within diameter
    neighbor = {}
    for n1 in nodeIDs:
        for n2 in nodeIDs:
            if (n1 != n2 and edges[n1, n2] <= diameter):
                if (n1 not in neighbor):
                    neighbor[n1] = []
                neighbor[n1].append(n2)

    # Find seed cliques in the neighbor graph =================================
    # FIXME: Now using stupid method, will be replaced by clique searching algorithm, or should I?
    # 3-clique
    for n1 in neighbor:
        for n2 in neighbor[n1]:
            for n3 in neighbor[n2]:
                if (n1 < n2 < n3 and n1 in neighbor[n3]):
                    seedClique.append([n1, n2, n3])

    # Try to union existing seed-clique to find larger clique =================
    canUnionFlag = True
    while (canUnionFlag):
        canUnionFlag = False
        for c1 in range(len(seedClique) - 1):
            for c2 in range(c1 + 1, len(seedClique)):
                # Try to merge two cliques
                clique1 = [i for i in seedClique[c1]]
                clique2 = [i for i in seedClique[c2]]
                intersect = listSetIntersect(clique1, clique2)                
                # Two cliques can be merged iff they have intersection
                if (len(intersect) > 0):
                    diff1 = listSetMinus(clique1, intersect)
                    diff2 = listSetMinus(clique2, intersect)                    
                    # Try to see if the nodes that are not in the intersection are close
                    mergeFlag = True
                    for n1 in diff1:
                        for n2 in diff2:
                            if (n1 not in neighbor[n2]):
                                mergeFlag = False
                                break
                        if (not mergeFlag):
                            break
                    if (mergeFlag):
                        newClique = listSetUnion(clique1, clique2)
                        newClique.sort()
                        seedClique.remove(clique1)
                        seedClique.remove(clique2)
                        seedClique.append(newClique)
                        canUnionFlag = True
                        break
            if (canUnionFlag):
                break
    return seedClique

def getSortedNodesByDist(
    nodes:      "Dictionary, returns the coordinate of given nodeID, \
                    {\
                        nodeID1: {'loc': (x, y)}, \
                        nodeID2: {'loc': (x, y)}, \
                        ... \
                    }" = None, 
    edges:      "1) String (default) 'Euclidean' or \
                 2) String 'LatLon'" = "Euclidean",
    nodeIDs:    "1) String (default) 'All', or \
                 2) A list of node IDs" = 'All',
    refNodeID:  "List, [x, y], the reference location to calculate distance" = None    
    ) -> "Given a set of locations, and a referencing node, sort the nodes by distance to this referencing node":

    # Define nodeIDs ==========================================================
    if (type(nodeIDs) is not list):
        if (nodeIDs == 'All'):
            nodeIDs = []
            for i in nodes:
                nodeIDs.append(i)

    # Define edges ============================================================
    if (type(edges) is not dict):
        if (edges == 'Euclidean'):
            edges = getTauEuclidean(nodes)
        elif (edges == 'LatLon'):
            edges = getTauLatLon(nodes)
        else:
            print("Error: Incorrect type `edges`")
            return None

    # Sort distance ===========================================================
    sortedSeq = []
    sortedSeqHeap = []
    for n in nodeIDs:
        dist = edges[refNodeID, n]
        heapq.heappush(sortedSeqHeap, (dist, n))
    while (len(sortedSeqHeap) > 0):
        sortedSeq.append(heapq.heappop(sortedSeqHeap)[1])  

    return sortedSeq

def getSweepSeq(
    nodes:      "Dictionary, returns the coordinate of given nodeID, \
                    {\
                        nodeID1: {'loc': (x, y)}, \
                        nodeID2: {'loc': (x, y)}, \
                        ... \
                    }" = None, 
    nodeIDs:    "1) String (default) 'All', or \
                 2) A list of node IDs" = 'All',
    centerLoc:  "1) (Default) String, 'centroid' the centroid of nodes, or\
                 2) List, [x, y], the center point" = None,
    isClockwise: "True if the sweeping direction is clock-wise, False otherwise" = True,
    initDeg:    "Starting direction of the sweeping, 0 as North" = 0
    ) -> "Given a set of locations, and a center point, gets the sequence from sweeping":

    # Define nodeIDs ==========================================================
    if (type(nodeIDs) is not list):
        if (nodeIDs == 'All'):
            nodeIDs = []
            for i in nodes:
                nodeIDs.append(i)

    # Initialize centroid =====================================================
    if (centerLoc == 'centroid'):
        centerLoc = getCentroid(nodes)

    # Initialize heap =========================================================
    degHeap = []
    centerLocNodes = []
    
    # Build heap ==============================================================
    for n in nodeIDs:
        dist = distEuclidean2D(nodes[n]['loc'], centerLoc)
        # If the nodes are too close, separate it/them
        if (dist <= CONST_EPSILON):
            centerLocNodes.append(n)
        else:
            dx = nodes[n]['loc'][0] - centerLoc[0]
            dy = nodes[n]['loc'][1] - centerLoc[1]
            (_, deg) = vecXY2Polar((dx, dy))
            # Calculate angles
            evalDeg = None
            if (isClockwise):
                evalDeg = deg - initDeg
            else:
                evalDeg = initDeg - deg
            while(evalDeg > 360):
                evalDeg -= 360
            while(evalDeg < 0):
                evalDeg += 360
            heapq.heappush(degHeap, (evalDeg, n))

    # Sweep ===================================================================
    sweepSeq = []
    while (len(degHeap)):
        sweepSeq.append(heapq.heappop(degHeap)[1])
    sweepSeq.extend(centerLocNodes)

    return sweepSeq

def getCentroid(
    nodes:      "Dictionary, returns the coordinate of given nodeID, \
                    {\
                        nodeID1: {'loc': (x, y)}, \
                        nodeID2: {'loc': (x, y)}, \
                        ... \
                    }" = None,
    nodeIDs:    "1) String (default) 'All', or \
                 2) A list of node IDs" = 'All',
    algo:       "1) String, 'Weiszfeld'" = 'Weiszfeld'
    ) -> "Get centroid location for given nodes":

    # Define nodeIDs ==========================================================
    if (type(nodeIDs) is not list):
        if (nodeIDs == 'All'):
            nodeIDs = []
            for i in nodes:
                nodeIDs.append(i)

    # Subroutine for finding centroid =========================================
    def _getCentroidWeiszfeld(nodes, nodeIDs):
        # Initialize ----------------------------------------------------------
        q = [1 for i in range(len(nodeIDs))]
        a = [nodes[nodeIDs[i]]['loc'][0] for i in range(len(nodeIDs))]
        b = [nodes[nodeIDs[i]]['loc'][1] for i in range(len(nodeIDs))]
        x = sum(a) / len(a)
        y = sum(b) / len(b)
        f = 0
        for i in range(len(nodeIDs)):
            f += math.sqrt((x - a[i])**2 + (y - b[i])**2)

        # Iterations ----------------------------------------------------------
        canGoFlag = True
        while (canGoFlag):
            # update q
            q = []
            for i in range(len(nodeIDs)):
                q.append(1 / math.sqrt((x - a[i])**2 + (y - b[i])**2))

            # update x, y
            x = 0
            y = 0
            for i in range(len(nodeIDs)):
                x += q[i] * a[i]
                y += q[i] * b[i]
            x /= sum(q)
            y /= sum(q)

            # update f
            newF = 0
            for i in range(len(nodeIDs)):
                newF += math.sqrt((x - a[i])**2 + (y - b[i])**2)
            if (abs(newF - f) < CONST_EPSILON):
                canGoFlag = False
            f = newF

        # Output --------------------------------------------------------------
        centroid = (x, y)

        return centroid

    # Call subroutines ========================================================
    centroid = None
    if (algo == "Weiszfeld"):
        centroid = _getCentroidWeiszfeld(nodes, nodeIDs)
    else:
        return None

    return centroid

def getScan(
    nodes:      "Dictionary, returns the coordinate of given nodeID, \
                    {\
                        nodeID1: {'loc': (x, y)}, \
                        nodeID2: {'loc': (x, y)}, \
                        ... \
                    }" = None,
    direction:  "Direction of scanning, 0 as North" = 0
    ) -> "Scan nodes from one direction":

    # Initialize ==============================================================
    baseline = []
    maxDist = 0
    centroid = getCentroid(nodes)
    for n in nodes:
        d = distEuclidean2D(nodes[n]['loc'], centroid)
        if (maxDist == None or d > maxDist):
            maxDist = 1.2 * d
    basePt = pointInDistXY(centroid, direction, maxDist)
    baseline = getPerpendicularLine(basePt, vecPolar2XY([10, direction]))

    # Distance to the baseline ================================================
    distHeap = []
    for n in nodes:
        dist2Baseline = distPt2Line(nodes[n]['loc'], baseline)
        heapq.heappush(distHeap, (dist2Baseline, n))

    # Scan ====================================================================
    scanSeq = []
    while (len(distHeap)):
        scanSeq.append(heapq.heappop(distHeap)[1])

    return scanSeq

def getConvexHull(
    nodes:      "Dictionary, returns the coordinate of given nodeID, \
                    {\
                        nodeID1: {'loc': (x, y)}, \
                        nodeID2: {'loc': (x, y)}, \
                        ... \
                    }" = None,
    algo:       "1) String, 'Jarvis', O(nH) (current implementation is O(nHlog n)) or\
                 2) String, (not available) 'DNC', O(nlog n) or\
                 3) String, (not available) 'Graham', O(nlog n) or\
                 4) String, (not available) 'Melkman'" = "Jarvis"
    ) -> "Given a set of node locations, return a list of nodeID which construct the convex hull":

    # Initialize ==============================================================
    chSeq = None

    # Some extreme cases ======================================================
    if (len(nodes) == 0):
        return None
    elif (len(nodes) <= 3):
        chSeq = []
        for n in nodes:
            chSeq.append(n)
        return chSeq

    # Subroutines for finding convex hull =====================================
    def _getConvexHullJavis():
        # References ----------------------------------------------------------
        # 1. https://blog.csdn.net/Bone_ACE/article/details/46239187
        # 2. chrome-extension://efaidnbmnnnibpcajpcglclefindmkaj/viewer.html?pdfurl=http%3A%2F%2Fwww.ams.sunysb.edu%2F~jsbm%2Fcourses%2F345%2F13%2Fmelkman.pdf&clen=46562&chunk=true
        # 3. https://en.wikipedia.org/wiki/Convex_hull_algorithms

        # Initialize ----------------------------------------------------------
        chSeq = []

        # Get the location of the left-most nodeID ----------------------------
        # Find an initial point which guaranteed to be in convex hull
        leftMostID = None
        leftMostX = None
        for n in nodes:
            if (leftMostID == None or nodes[n]['loc'][0] < leftMostX):
                leftMostID = n
                leftMostX = nodes[n]['loc'][0]

        # Jarvis march --------------------------------------------------------
        curNodeID = leftMostID
        curDirection = 0
        marchFlag = True
        while (marchFlag):
            sweepSeq = getSweepSeq(
                nodes = nodes,
                nodeIDs = [i for i in nodes if i not in chSeq],
                centerLoc = nodes[curNodeID]['loc'],
                initDeg = curDirection)
            if (sweepSeq[0] == leftMostID):
                marchFlag = False
            chSeq.append(sweepSeq[0])
            curDirection = getHeadingXY(nodes[curNodeID]['loc'], nodes[sweepSeq[0]]['loc'])    
            curNodeID = sweepSeq[0]
        return chSeq

    # Call subroutines ========================================================
    if (algo == "Jarvis"):
        chSeq = _getConvexHullJavis()
    else:
        return None
    
    return chSeq
